@@ -18,7 +18,7 @@ from .exceptions import (
     InferredBucketConflict, CannotInferEnvFromNoGlobalEnvs, CannotInferEnvFromManyGlobalEnvs, MissingGlobalEnv,
     GlobalBucketAccessError, SynonymousEnvironmentVariablesMismatched,
 )
-from .misc_utils import PRINT, override_environ, ignored, exported
+from .misc_utils import PRINT, override_environ, ignored, exported, merge_key_value_dict_lists, key_value_dict
 
 
 exported(get_beanstalk_real_url)
@@ -50,6 +50,8 @@ class HealthPageKey:  # This is moving here from cgap-portal.
     NAMESPACE = 'namespace'
     PROCESSED_FILE_BUCKET = 'processed_file_bucket'          # = s3Utils.OUTFILE_BUCKET_HEALTH_PAGE_KEY
     PROJECT_VERSION = 'project_version'
+    PYTHON_VERSION = 'python_version'
+    S3_ENCRYPT_KEY_ID = 's3_encrypt_key_id'
     SNOVAULT_VERSION = 'snovault_version'
     SYSTEM_BUCKET = 'system_bucket'                          # = s3Utils.SYS_BUCKET_HEALTH_PAGE_KEY
     TIBANNA_CWLS_BUCKET = 'tibanna_cwls_bucket'              # = s3Utils.TIBANNA_CWLS_BUCKET_HEALTH_PAGE_KEY
@@ -70,6 +72,8 @@ class s3Utils(object):  # NOQA - This class name violates style rules, but a lot
     TIBANNA_OUTPUT_BUCKET_SUFFIX = 'tibanna-output'
     TIBANNA_CWLS_BUCKET_SUFFIX = 'tibanna-cwls'
 
+    s3_encrypt_key_id = None  # default. might be overridden based on health page in various places below
+
     EB_PREFIX = "elasticbeanstalk"
     EB_AND_ENV_PREFIX = EB_PREFIX + "-%s-"  # = "elasticbeanstalk-%s-"
 
@@ -131,6 +135,7 @@ class s3Utils(object):  # NOQA - This class name violates style rules, but a lot
                 health_json_url = f'{global_manager.portal_url}/health?format=json'
                 logger.warning('health json url: {}'.format(health_json_url))
                 health_json = EnvManager.fetch_health_page_json(url=health_json_url)
+                self.s3_encrypt_key_id = health_json.get(HealthPageKey.S3_ENCRYPT_KEY_ID, None)
                 sys_bucket_from_health_page = health_json[HealthPageKey.SYSTEM_BUCKET]
                 outfile_bucket_from_health_page = health_json[HealthPageKey.PROCESSED_FILE_BUCKET]
                 raw_file_bucket_from_health_page = health_json[HealthPageKey.FILE_UPLOAD_BUCKET]
@@ -200,6 +205,7 @@ class s3Utils(object):  # NOQA - This class name violates style rules, but a lot
                     if not es_url.startswith("http"):  # will match http: and https:
                         es_url = f"https://{es_url}"
                     self.env_manager = EnvManager.compose(portal_url=self.url, es_url=es_url, env_name=env, s3=self.s3)
+                    self.s3_encrypt_key_id = health_json.get(HealthPageKey.S3_ENCRYPT_KEY_ID, None)
 
                 # TODO: This branch is not setting self.global_env_bucket_manager, but it _could_ do that from the
                 #       description. -kmp 21-Aug-2021
@@ -287,6 +293,83 @@ class s3Utils(object):  # NOQA - This class name violates style rules, but a lot
             return False
         return file_metadata
 
+    def get_object_tags(self, *, bucket, key):
+        """
+        Get all tags of an object.
+
+        Args:
+            bucket (string): S3 bucket
+            key (string): object key
+
+        Returns:
+            Returns list of object tags.
+            The format is [{'Key': 'KEY1','Value': 'VALUE1'},{...}]
+        """
+
+        try:
+            response = self.s3.get_object_tagging(
+                Bucket=bucket,
+                Key=key,
+            )
+            return response["TagSet"]
+        except Exception as e:
+            logger.warning(f'Could not get tags for object {bucket}/{key}: {str(e)}')
+            raise e
+
+    def set_object_tag(self, *, bucket, key, tag_key, tag_value):
+        """
+        Adds (tag_key,tag_value) pair to the object. If a tag with key tag_key is already present,
+        it will be overwritten.
+
+        Args:
+            bucket (string): S3 bucket
+            key (string): object key
+            tag_key (string): Tag key
+            tag_value (string): Tag value
+
+        Returns:
+            Dict with the versionId of the object the tag-set was added to
+        """
+
+        return self.set_object_tags(
+            bucket=bucket,
+            key=key,
+            tags=[key_value_dict(tag_key, tag_value)],
+            merge_existing_tags=True,
+        )
+
+    def set_object_tags(self, *, bucket, key, tags, merge_existing_tags=True):
+        """
+        Adds or replaces tags of an object with the ones specified in `tags`.
+
+        Args:
+            bucket (string): S3 bucket
+            key (string): object key
+            tags (list): List of tags of the form [{'Key': 'KEY1','Value': 'VALUE1'}, {...}]
+            merge_existing_tags (bool): If False, existing tags are replaced with the provided ones (use with care!).
+                Otherwise, the provided tags are merged with the existing ones (default)
+
+        Returns:
+            Dict with the versionId of the object the tag-set was added to
+        """
+
+        try:
+            new_tags = tags
+            if merge_existing_tags:
+                existing_tags = self.get_object_tags(bucket=bucket, key=key)
+                if existing_tags:
+                    new_tags = merge_key_value_dict_lists(existing_tags, new_tags)
+            return self.s3.put_object_tagging(
+                Bucket=bucket,
+                Key=key,
+                Tagging={
+                    'TagSet': new_tags,
+                },
+            )
+        except Exception as e:
+            logger.warning(f'{bucket}/{key} could not be tagged: {str(e)}')
+            raise e
+
     def get_file_size(self, key, bucket=None, add_bytes=0, add_gb=0,
                       size_in_gb=False):
         """
